@@ -2,7 +2,7 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
-use compact_cow_str::CompactCowStr;
+use cow_rc_str::CowRcStr;
 use std::ops::Range;
 use std::ascii::AsciiExt;
 use std::ops::BitOr;
@@ -28,7 +28,7 @@ pub enum BasicParseError<'a> {
     /// The end of the input was encountered unexpectedly.
     EndOfInput,
     /// An `@` rule was encountered that was invalid.
-    AtRuleInvalid(CompactCowStr<'a>),
+    AtRuleInvalid(CowRcStr<'a>),
     /// The body of an '@' rule was invalid.
     AtRuleBodyInvalid,
     /// A qualified rule was encountered that was invalid.
@@ -61,12 +61,29 @@ impl<'a, T> ParseError<'a, T> {
 }
 
 /// The owned input for a parser.
-pub struct ParserInput<'t>(Tokenizer<'t>);
+pub struct ParserInput<'i> {
+    tokenizer: Tokenizer<'i>,
+    cached_token: Option<CachedToken<'i>>,
+}
 
-impl<'t> ParserInput<'t> {
+struct CachedToken<'i> {
+    token: Token<'i>,
+    start_position: tokenizer::SourcePosition,
+    end_position: tokenizer::SourcePosition,
+}
+
+impl<'i> ParserInput<'i> {
     /// Create a new input for a parser.
-    pub fn new(input: &'t str) -> ParserInput<'t> {
-        ParserInput(Tokenizer::new(input))
+    pub fn new(input: &'i str) -> ParserInput<'i> {
+        ParserInput {
+            tokenizer: Tokenizer::new(input),
+            cached_token: None,
+        }
+    }
+
+    #[inline]
+    fn cached_token_ref(&self) -> &Token<'i> {
+        &self.cached_token.as_ref().unwrap().token
     }
 }
 
@@ -74,7 +91,7 @@ impl<'t> ParserInput<'t> {
 /// yields `Token`s,
 /// and keeps track of nested blocks and functions.
 pub struct Parser<'i: 't, 't> {
-    tokenizer: &'t mut ParserInput<'i>,
+    input: &'t mut ParserInput<'i>,
     /// If `Some(_)`, .parse_nested_block() can be called.
     at_start_of: Option<BlockType>,
     /// For parsers from `parse_until` or `parse_nested_block`
@@ -182,7 +199,7 @@ impl<'i: 't, 't> Parser<'i, 't> {
     #[inline]
     pub fn new(input: &'t mut ParserInput<'i>) -> Parser<'i, 't> {
         Parser {
-            tokenizer: input,
+            input: input,
             at_start_of: None,
             stop_before: Delimiter::None,
         }
@@ -190,7 +207,7 @@ impl<'i: 't, 't> Parser<'i, 't> {
 
     /// Return the current line that is being parsed.
     pub fn current_line(&self) -> &'i str {
-        self.tokenizer.0.current_source_line()
+        self.input.tokenizer.current_source_line()
     }
 
     /// Check whether the input is exhausted. That is, if `.next()` would return a token.
@@ -211,7 +228,7 @@ impl<'i: 't, 't> Parser<'i, 't> {
         let result = match self.next() {
             Err(BasicParseError::EndOfInput) => Ok(()),
             Err(e) => unreachable!("Unexpected error encountered: {:?}", e),
-            Ok(t) => Err(BasicParseError::UnexpectedToken(t)),
+            Ok(t) => Err(BasicParseError::UnexpectedToken(t.clone())),
         };
         self.reset(start_position);
         result
@@ -223,7 +240,7 @@ impl<'i: 't, 't> Parser<'i, 't> {
     #[inline]
     pub fn position(&self) -> SourcePosition {
         SourcePosition {
-            position: (self.tokenizer.0).position(),
+            position: self.input.tokenizer.position(),
             at_start_of: self.at_start_of,
         }
     }
@@ -234,35 +251,35 @@ impl<'i: 't, 't> Parser<'i, 't> {
     /// Should only be used with `SourcePosition` values from the same `Parser` instance.
     #[inline]
     pub fn reset(&mut self, new_position: SourcePosition) {
-        (self.tokenizer.0).reset(new_position.position);
+        self.input.tokenizer.reset(new_position.position);
         self.at_start_of = new_position.at_start_of;
     }
 
     /// Start looking for `var()` functions. (See the `.seen_var_functions()` method.)
     #[inline]
     pub fn look_for_var_functions(&mut self) {
-        (self.tokenizer.0).look_for_var_functions()
+        self.input.tokenizer.look_for_var_functions()
     }
 
     /// Return whether a `var()` function has been seen by the tokenizer since
     /// either `look_for_var_functions` was called, and stop looking.
     #[inline]
     pub fn seen_var_functions(&mut self) -> bool {
-        (self.tokenizer.0).seen_var_functions()
+        self.input.tokenizer.seen_var_functions()
     }
 
     /// Start looking for viewport percentage lengths. (See the `seen_viewport_percentages`
     /// method.)
     #[inline]
     pub fn look_for_viewport_percentages(&mut self) {
-        (self.tokenizer.0).look_for_viewport_percentages()
+        self.input.tokenizer.look_for_viewport_percentages()
     }
 
     /// Return whether a `vh`, `vw`, `vmin`, or `vmax` dimension has been seen by the tokenizer
     /// since `look_for_viewport_percentages` was called, and stop looking.
     #[inline]
     pub fn seen_viewport_percentages(&mut self) -> bool {
-        (self.tokenizer.0).seen_viewport_percentages()
+        self.input.tokenizer.seen_viewport_percentages()
     }
 
     /// Execute the given closure, passing it the parser.
@@ -283,25 +300,25 @@ impl<'i: 't, 't> Parser<'i, 't> {
     /// Return a slice of the CSS input
     #[inline]
     pub fn slice(&self, range: Range<SourcePosition>) -> &'i str {
-        (self.tokenizer.0).slice(range.start.position..range.end.position)
+        self.input.tokenizer.slice(range.start.position..range.end.position)
     }
 
     /// Return a slice of the CSS input, from the given position to the current one.
     #[inline]
     pub fn slice_from(&self, start_position: SourcePosition) -> &'i str {
-        (self.tokenizer.0).slice_from(start_position.position)
+        self.input.tokenizer.slice_from(start_position.position)
     }
 
     /// Return the line and column number within the input for the current position.
     #[inline]
     pub fn current_source_location(&self) -> SourceLocation {
-        (self.tokenizer.0).current_source_location()
+        self.input.tokenizer.current_source_location()
     }
 
     /// Return the line and column number within the input for the given position.
     #[inline]
     pub fn source_location(&self, target: SourcePosition) -> SourceLocation {
-        (self.tokenizer.0).source_location(target.position)
+        self.input.tokenizer.source_location(target.position)
     }
 
     /// Return the next token in the input that is neither whitespace or a comment,
@@ -315,23 +332,27 @@ impl<'i: 't, 't> Parser<'i, 't> {
     /// See the `Parser::parse_nested_block` method to parse the content of functions or blocks.
     ///
     /// This only returns a closing token when it is unmatched (and therefore an error).
-    pub fn next(&mut self) -> Result<Token<'i>, BasicParseError<'i>> {
+    pub fn next(&mut self) -> Result<&Token<'i>, BasicParseError<'i>> {
         loop {
             match self.next_including_whitespace_and_comments() {
-                Ok(Token::WhiteSpace(_)) | Ok(Token::Comment(_)) => {},
-                result => return result
+                Err(e) => return Err(e),
+                Ok(&Token::WhiteSpace(_)) | Ok(&Token::Comment(_)) => {},
+                _ => break
             }
         }
+        Ok(self.input.cached_token_ref())
     }
 
     /// Same as `Parser::next`, but does not skip whitespace tokens.
-    pub fn next_including_whitespace(&mut self) -> Result<Token<'i>, BasicParseError<'i>> {
+    pub fn next_including_whitespace(&mut self) -> Result<&Token<'i>, BasicParseError<'i>> {
         loop {
             match self.next_including_whitespace_and_comments() {
-                Ok(Token::Comment(_)) => {},
-                result => return result
+                Err(e) => return Err(e),
+                Ok(&Token::Comment(_)) => {},
+                _ => break
             }
         }
+        Ok(self.input.cached_token_ref())
     }
 
     /// Same as `Parser::next`, but does not skip whitespace or comment tokens.
@@ -340,16 +361,40 @@ impl<'i: 't, 't> Parser<'i, 't> {
     /// where comments are preserved.
     /// When parsing higher-level values, per the CSS Syntax specification,
     /// comments should always be ignored between tokens.
-    pub fn next_including_whitespace_and_comments(&mut self) -> Result<Token<'i>, BasicParseError<'i>> {
+    pub fn next_including_whitespace_and_comments(&mut self) -> Result<&Token<'i>, BasicParseError<'i>> {
         if let Some(block_type) = self.at_start_of.take() {
-            consume_until_end_of_block(block_type, &mut self.tokenizer.0);
+            consume_until_end_of_block(block_type, &mut self.input.tokenizer);
         }
-        let byte = (self.tokenizer.0).next_byte();
+
+        let byte = self.input.tokenizer.next_byte();
         if self.stop_before.contains(Delimiters::from_byte(byte)) {
             return Err(BasicParseError::EndOfInput)
         }
-        let token = (self.tokenizer.0).next().map_err(|()| BasicParseError::EndOfInput)?;
-        if let Some(block_type) = BlockType::opening(&token) {
+
+        let token_start_position = self.input.tokenizer.position();
+        let token;
+        match self.input.cached_token {
+            Some(ref cached_token) if cached_token.start_position == token_start_position => {
+                self.input.tokenizer.reset(cached_token.end_position);
+                match cached_token.token {
+                    Token::Dimension { ref unit, .. } => self.input.tokenizer.see_dimension(unit),
+                    Token::Function(ref name) => self.input.tokenizer.see_function(name),
+                    _ => {}
+                }
+                token = &cached_token.token
+            }
+            _ => {
+                let new_token = self.input.tokenizer.next().map_err(|()| BasicParseError::EndOfInput)?;
+                self.input.cached_token = Some(CachedToken {
+                    token: new_token,
+                    start_position: token_start_position,
+                    end_position: self.input.tokenizer.position(),
+                });
+                token = self.input.cached_token_ref()
+            }
+        }
+
+        if let Some(block_type) = BlockType::opening(token) {
             self.at_start_of = Some(block_type);
         }
         Ok(token)
@@ -385,7 +430,7 @@ impl<'i: 't, 't> Parser<'i, 't> {
             values.push(self.parse_until_before(Delimiter::Comma, &mut parse_one)?);
             match self.next() {
                 Err(_) => return Ok(values),
-                Ok(Token::Comma) => continue,
+                Ok(&Token::Comma) => continue,
                 Ok(_) => unreachable!(),
             }
         }
@@ -438,94 +483,105 @@ impl<'i: 't, 't> Parser<'i, 't> {
     /// Parse a <whitespace-token> and return its value.
     #[inline]
     pub fn expect_whitespace(&mut self) -> Result<&'i str, BasicParseError<'i>> {
-        match self.next_including_whitespace()? {
+        match *self.next_including_whitespace()? {
             Token::WhiteSpace(value) => Ok(value),
-            t => Err(BasicParseError::UnexpectedToken(t))
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
     /// Parse a <ident-token> and return the unescaped value.
     #[inline]
-    pub fn expect_ident(&mut self) -> Result<CompactCowStr<'i>, BasicParseError<'i>> {
-        match self.next()? {
-            Token::Ident(value) => Ok(value),
-            t => Err(BasicParseError::UnexpectedToken(t))
+    pub fn expect_ident(&mut self) -> Result<&CowRcStr<'i>, BasicParseError<'i>> {
+        match *self.next()? {
+            Token::Ident(ref value) => Ok(value),
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
+    /// expect_ident, but clone the CowRcStr
+    #[inline]
+    pub fn expect_ident_cloned(&mut self) -> Result<CowRcStr<'i>, BasicParseError<'i>> {
+        self.expect_ident().map(|s| s.clone())
+    }
+
     /// Parse a <ident-token> whose unescaped value is an ASCII-insensitive match for the given value.
     #[inline]
     pub fn expect_ident_matching(&mut self, expected_value: &str) -> Result<(), BasicParseError<'i>> {
-        match self.next()? {
+        match *self.next()? {
             Token::Ident(ref value) if value.eq_ignore_ascii_case(expected_value) => Ok(()),
-            t => Err(BasicParseError::UnexpectedToken(t))
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
     /// Parse a <string-token> and return the unescaped value.
     #[inline]
-    pub fn expect_string(&mut self) -> Result<CompactCowStr<'i>, BasicParseError<'i>> {
-        match self.next()? {
-            Token::QuotedString(value) => Ok(value),
-            t => Err(BasicParseError::UnexpectedToken(t))
+    pub fn expect_string(&mut self) -> Result<&CowRcStr<'i>, BasicParseError<'i>> {
+        match *self.next()? {
+            Token::QuotedString(ref value) => Ok(value),
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
+    /// expect_string, but clone the CowRcStr
+    #[inline]
+    pub fn expect_string_cloned(&mut self) -> Result<CowRcStr<'i>, BasicParseError<'i>> {
+        self.expect_string().map(|s| s.clone())
+    }
+
     /// Parse either a <ident-token> or a <string-token>, and return the unescaped value.
     #[inline]
-    pub fn expect_ident_or_string(&mut self) -> Result<CompactCowStr<'i>, BasicParseError<'i>> {
-        match self.next()? {
-            Token::Ident(value) => Ok(value),
-            Token::QuotedString(value) => Ok(value),
-            t => Err(BasicParseError::UnexpectedToken(t))
+    pub fn expect_ident_or_string(&mut self) -> Result<&CowRcStr<'i>, BasicParseError<'i>> {
+        match *self.next()? {
+            Token::Ident(ref value) => Ok(value),
+            Token::QuotedString(ref value) => Ok(value),
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
     /// Parse a <url-token> and return the unescaped value.
     #[inline]
-    pub fn expect_url(&mut self) -> Result<CompactCowStr<'i>, BasicParseError<'i>> {
-        match self.next()? {
-            Token::UnquotedUrl(value) => Ok(value),
-            Token::Function(ref name) if name.eq_ignore_ascii_case("url") => {
-                self.parse_nested_block(|input| input.expect_string().map_err(ParseError::Basic))
-                    .map_err(ParseError::<()>::basic)
-            },
-            t => Err(BasicParseError::UnexpectedToken(t))
+    pub fn expect_url(&mut self) -> Result<CowRcStr<'i>, BasicParseError<'i>> {
+        // FIXME: revert early returns when lifetimes are non-lexical
+        match *self.next()? {
+            Token::UnquotedUrl(ref value) => return Ok(value.clone()),
+            Token::Function(ref name) if name.eq_ignore_ascii_case("url") => {}
+            ref t => return Err(BasicParseError::UnexpectedToken(t.clone()))
         }
+        self.parse_nested_block(|input| input.expect_string().map_err(ParseError::Basic).map(|s| s.clone()))
+            .map_err(ParseError::<()>::basic)
     }
 
     /// Parse either a <url-token> or a <string-token>, and return the unescaped value.
     #[inline]
-    pub fn expect_url_or_string(&mut self) -> Result<CompactCowStr<'i>, BasicParseError<'i>> {
-        match self.next()? {
-            Token::UnquotedUrl(value) => Ok(value),
-            Token::QuotedString(value) => Ok(value),
-            Token::Function(ref name) if name.eq_ignore_ascii_case("url") => {
-                self.parse_nested_block(|input| input.expect_string().map_err(ParseError::Basic))
-                    .map_err(ParseError::<()>::basic)
-            },
-            t => Err(BasicParseError::UnexpectedToken(t))
+    pub fn expect_url_or_string(&mut self) -> Result<CowRcStr<'i>, BasicParseError<'i>> {
+        // FIXME: revert early returns when lifetimes are non-lexical
+        match *self.next()? {
+            Token::UnquotedUrl(ref value) => return Ok(value.clone()),
+            Token::QuotedString(ref value) => return Ok(value.clone()),
+            Token::Function(ref name) if name.eq_ignore_ascii_case("url") => {}
+            ref t => return Err(BasicParseError::UnexpectedToken(t.clone()))
         }
+        self.parse_nested_block(|input| input.expect_string().map_err(ParseError::Basic).map(|s| s.clone()))
+            .map_err(ParseError::<()>::basic)
     }
 
     /// Parse a <number-token> and return the integer value.
     #[inline]
     pub fn expect_number(&mut self) -> Result<f32, BasicParseError<'i>> {
-        match self.next()? {
+        match *self.next()? {
             Token::Number { value, .. } => Ok(value),
-            t => Err(BasicParseError::UnexpectedToken(t))
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
     /// Parse a <number-token> that does not have a fractional part, and return the integer value.
     #[inline]
     pub fn expect_integer(&mut self) -> Result<i32, BasicParseError<'i>> {
-        let token = self.next()?;
-        match token {
+        match *self.next()? {
             Token::Number { int_value: Some(int_value), .. } => {
                 Ok(int_value)
             }
-            t => Err(BasicParseError::UnexpectedToken(t))
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
@@ -533,45 +589,45 @@ impl<'i: 't, 't> Parser<'i, 't> {
     /// `0%` and `100%` map to `0.0` and `1.0` (not `100.0`), respectively.
     #[inline]
     pub fn expect_percentage(&mut self) -> Result<f32, BasicParseError<'i>> {
-        match self.next()? {
+        match *self.next()? {
             Token::Percentage { unit_value, .. } => Ok(unit_value),
-            t => Err(BasicParseError::UnexpectedToken(t))
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
     /// Parse a `:` <colon-token>.
     #[inline]
     pub fn expect_colon(&mut self) -> Result<(), BasicParseError<'i>> {
-        match self.next()? {
+        match *self.next()? {
             Token::Colon => Ok(()),
-            t => Err(BasicParseError::UnexpectedToken(t))
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
     /// Parse a `;` <semicolon-token>.
     #[inline]
     pub fn expect_semicolon(&mut self) -> Result<(), BasicParseError<'i>> {
-        match self.next()? {
+        match *self.next()? {
             Token::Semicolon => Ok(()),
-            t => Err(BasicParseError::UnexpectedToken(t))
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
     /// Parse a `,` <comma-token>.
     #[inline]
     pub fn expect_comma(&mut self) -> Result<(), BasicParseError<'i>> {
-        match self.next()? {
+        match *self.next()? {
             Token::Comma => Ok(()),
-            t => Err(BasicParseError::UnexpectedToken(t))
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
     /// Parse a <delim-token> with the given value.
     #[inline]
     pub fn expect_delim(&mut self, expected_value: char) -> Result<(), BasicParseError<'i>> {
-        match self.next()? {
+        match *self.next()? {
             Token::Delim(value) if value == expected_value => Ok(()),
-            t => Err(BasicParseError::UnexpectedToken(t))
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
@@ -580,9 +636,9 @@ impl<'i: 't, 't> Parser<'i, 't> {
     /// If the result is `Ok`, you can then call the `Parser::parse_nested_block` method.
     #[inline]
     pub fn expect_curly_bracket_block(&mut self) -> Result<(), BasicParseError<'i>> {
-        match self.next()? {
+        match *self.next()? {
             Token::CurlyBracketBlock => Ok(()),
-            t => Err(BasicParseError::UnexpectedToken(t))
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
@@ -591,9 +647,9 @@ impl<'i: 't, 't> Parser<'i, 't> {
     /// If the result is `Ok`, you can then call the `Parser::parse_nested_block` method.
     #[inline]
     pub fn expect_square_bracket_block(&mut self) -> Result<(), BasicParseError<'i>> {
-        match self.next()? {
+        match *self.next()? {
             Token::SquareBracketBlock => Ok(()),
-            t => Err(BasicParseError::UnexpectedToken(t))
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
@@ -602,9 +658,9 @@ impl<'i: 't, 't> Parser<'i, 't> {
     /// If the result is `Ok`, you can then call the `Parser::parse_nested_block` method.
     #[inline]
     pub fn expect_parenthesis_block(&mut self) -> Result<(), BasicParseError<'i>> {
-        match self.next()? {
+        match *self.next()? {
             Token::ParenthesisBlock => Ok(()),
-            t => Err(BasicParseError::UnexpectedToken(t))
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
@@ -612,10 +668,10 @@ impl<'i: 't, 't> Parser<'i, 't> {
     ///
     /// If the result is `Ok`, you can then call the `Parser::parse_nested_block` method.
     #[inline]
-    pub fn expect_function(&mut self) -> Result<CompactCowStr<'i>, BasicParseError<'i>> {
-        match self.next()? {
-            Token::Function(name) => Ok(name),
-            t => Err(BasicParseError::UnexpectedToken(t))
+    pub fn expect_function(&mut self) -> Result<&CowRcStr<'i>, BasicParseError<'i>> {
+        match *self.next()? {
+            Token::Function(ref name) => Ok(name),
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
@@ -624,9 +680,9 @@ impl<'i: 't, 't> Parser<'i, 't> {
     /// If the result is `Ok`, you can then call the `Parser::parse_nested_block` method.
     #[inline]
     pub fn expect_function_matching(&mut self, expected_name: &str) -> Result<(), BasicParseError<'i>> {
-        match self.next()? {
+        match *self.next()? {
             Token::Function(ref name) if name.eq_ignore_ascii_case(expected_name) => Ok(()),
-            t => Err(BasicParseError::UnexpectedToken(t))
+            ref t => Err(BasicParseError::UnexpectedToken(t.clone()))
         }
     }
 
@@ -637,20 +693,22 @@ impl<'i: 't, 't> Parser<'i, 't> {
     pub fn expect_no_error_token(&mut self) -> Result<(), BasicParseError<'i>> {
         loop {
             match self.next_including_whitespace_and_comments() {
-                Ok(Token::Function(_)) | Ok(Token::ParenthesisBlock) |
-                Ok(Token::SquareBracketBlock) | Ok(Token::CurlyBracketBlock) => {
-                    let result = self.parse_nested_block(|input| input.expect_no_error_token()
-                                                         .map_err(|e| ParseError::Basic(e)));
-                    result.map_err(ParseError::<()>::basic)?
-                }
+                Ok(&Token::Function(_)) |
+                Ok(&Token::ParenthesisBlock) |
+                Ok(&Token::SquareBracketBlock) |
+                Ok(&Token::CurlyBracketBlock) => {}
                 Ok(token) => {
                     if token.is_parse_error() {
                         //FIXME: maybe these should be separate variants of BasicParseError instead?
-                        return Err(BasicParseError::UnexpectedToken(token))
+                        return Err(BasicParseError::UnexpectedToken(token.clone()))
                     }
+                    continue
                 }
                 Err(_) => return Ok(())
             }
+            let result = self.parse_nested_block(|input| input.expect_no_error_token()
+                                                 .map_err(|e| ParseError::Basic(e)));
+            result.map_err(ParseError::<()>::basic)?
         }
     }
 }
@@ -665,23 +723,23 @@ pub fn parse_until_before<'i: 't, 't, F, T, E>(parser: &mut Parser<'i, 't>,
     // Introduce a new scope to limit duration of nested_parser’s borrow
     {
         let mut delimited_parser = Parser {
-            tokenizer: parser.tokenizer,
+            input: parser.input,
             at_start_of: parser.at_start_of.take(),
             stop_before: delimiters,
         };
         result = delimited_parser.parse_entirely(parse);
         if let Some(block_type) = delimited_parser.at_start_of {
-            consume_until_end_of_block(block_type, &mut delimited_parser.tokenizer.0);
+            consume_until_end_of_block(block_type, &mut delimited_parser.input.tokenizer);
         }
     }
     // FIXME: have a special-purpose tokenizer method for this that does less work.
     loop {
-        if delimiters.contains(Delimiters::from_byte((parser.tokenizer.0).next_byte())) {
+        if delimiters.contains(Delimiters::from_byte((parser.input.tokenizer).next_byte())) {
             break
         }
-        if let Ok(token) = (parser.tokenizer.0).next() {
+        if let Ok(token) = (parser.input.tokenizer).next() {
             if let Some(block_type) = BlockType::opening(&token) {
-                consume_until_end_of_block(block_type, &mut parser.tokenizer.0);
+                consume_until_end_of_block(block_type, &mut parser.input.tokenizer);
             }
         } else {
             break
@@ -696,12 +754,12 @@ pub fn parse_until_after<'i: 't, 't, F, T, E>(parser: &mut Parser<'i, 't>,
                                               -> Result <T, ParseError<'i, E>>
     where F: for<'tt> FnOnce(&mut Parser<'i, 'tt>) -> Result<T, ParseError<'i, E>> {
     let result = parser.parse_until_before(delimiters, parse);
-    let next_byte = (parser.tokenizer.0).next_byte();
+    let next_byte = (parser.input.tokenizer).next_byte();
     if next_byte.is_some() && !parser.stop_before.contains(Delimiters::from_byte(next_byte)) {
         debug_assert!(delimiters.contains(Delimiters::from_byte(next_byte)));
-        (parser.tokenizer.0).advance(1);
+        (parser.input.tokenizer).advance(1);
         if next_byte == Some(b'{') {
-            consume_until_end_of_block(BlockType::CurlyBracket, &mut parser.tokenizer.0);
+            consume_until_end_of_block(BlockType::CurlyBracket, &mut parser.input.tokenizer);
         }
     }
     result
@@ -724,16 +782,16 @@ pub fn parse_nested_block<'i: 't, 't, F, T, E>(parser: &mut Parser<'i, 't>, pars
     // Introduce a new scope to limit duration of nested_parser’s borrow
     {
         let mut nested_parser = Parser {
-            tokenizer: parser.tokenizer,
+            input: parser.input,
             at_start_of: None,
             stop_before: closing_delimiter,
         };
         result = nested_parser.parse_entirely(parse);
         if let Some(block_type) = nested_parser.at_start_of {
-            consume_until_end_of_block(block_type, &mut nested_parser.tokenizer.0);
+            consume_until_end_of_block(block_type, &mut nested_parser.input.tokenizer);
         }
     }
-    consume_until_end_of_block(block_type, &mut parser.tokenizer.0);
+    consume_until_end_of_block(block_type, &mut parser.input.tokenizer);
     result
 }
 
